import io
import os
import sys
from keras.models import load_model
import numpy as np


def get_dir_from_file(filename):
    file = io.open(filename, "r", encoding="utf8")
    dir = {}
    dir["__NO_WORD__"] = 0
    i = 1
    for line in file:
        dir[line[:-1]] = i
        i += 1
    dir["__UNKNOWN__"] = i
    return dir


def get_index(w, vocab):
    if w in vocab:
        return vocab[w]
    return vocab["__UNKNOWN__"]

def get_set_from_file(filename):
    file = io.open(filename, "r", encoding="utf8")
    list = set()
    for line in file:
        line = line.replace("\n", '').split()
        list.add(line[0])
    return list


occ = {}
num = {}
denum = {}
gigasenses = ["Living_Animate_Entity", "Natural_Object", "Manufactured_Object", "Informational_Object",
              "Dynamic_situation", "Stative_situation"]


vocab_words = get_dir_from_file("../../dicts/words")
vocab_lemmas = get_dir_from_file("../../dicts/lemmas")
vocab_morphos = get_dir_from_file("../../dicts/morphos")
position_size = 300
nouns_finished = set()


def print_usage_and_exit():
    print("Usage: " + sys.argv[0] + "<nounsList> <fileName> ")
    sys.exit(0)

def add_padding(words):
    max_len = 0
    for word in words:
        if max_len < len(word):
            max_len = len(word)
    if max_len == 0:
        max_len = 1
    for i in range(len(words)):
        for j in range(len(words[i]), max_len):
            words[i].insert(0, 0)
    for word in words:
        if len(word) != max_len:
            print("BUG")


def update(buffer, nouns, result_file):
    current_words = []
    current_lemmas = []
    current_morphos = []
    current_positions = []
    word_positions = []
    words_left = []
    lemmas_left = []
    morphos_left = []
    words_right = []
    lemmas_right = []
    morphos_right = []
    morpho = []
    n = []
    current_nouns = []
    p = -1

    for line in buffer.split("\n"):
        line = line.split("\t")
        if len(line) != 7:
            continue
        p += 1
        current_words.append(get_index(line[0], vocab_words))
        current_lemmas.append(get_index(line[3], vocab_lemmas))
        current_morphos.append(get_index(line[2], vocab_morphos))
        current_positions.append(p)
        if line[1] == "NOUN" and line[3] in nouns:
            word_positions.append(p)
            current_nouns.append(line[3])
        if line[6] == "1":
            for i in range(len(word_positions)):
                new_positions_left = []
                new_positions_right = []
                for p in current_positions:
                    if p < word_positions[i]:
                        new_positions_left.append(abs(p - word_positions[i]))
                    if p > word_positions[i]:
                        new_positions_right.append(abs(p - word_positions[i]))

                if len(new_positions_left) >= position_size or len(new_positions_right) >= position_size:
                    continue
                n.append(current_nouns[i])
                words_left.append(current_words[:word_positions[i]])
                lemmas_left.append(current_lemmas[:word_positions[i]])
                morphos_left.append(current_morphos[:word_positions[i]])

                words_right.append(list(reversed(current_words[word_positions[i] + 1:])))
                lemmas_right.append(list(reversed(current_lemmas[word_positions[i] + 1:])))
                morphos_right.append(list(reversed(current_morphos[word_positions[i] + 1:])))
                morpho.append([current_morphos[word_positions[i]]])
            word_positions = []
            current_words = []
            current_lemmas = []
            current_morphos = []
            current_positions = []
            p = -1
    if words_left:
        add_padding(words_left)
        add_padding(words_right)
        add_padding(lemmas_right)
        add_padding(lemmas_left)
        add_padding(morphos_right)
        add_padding(morphos_left)
        add_padding(morpho)
        X = [np.array(words_left), np.array(lemmas_left), np.array(morphos_left),
             np.array(words_right), np.array(lemmas_right), np.array(morphos_right),
             np.array(morpho)]

        predictions = {}
        for g in gigasenses:
            predictions[g] = models[g].predict(X)
        for i in range(len(n)):
            result_file.write(n[i] + "\t")
            for g in gigasenses:
                prediction = predictions[g][i][0]
                result_file.write(str(prediction) + "\t")
            result_file.write("\n")


if __name__ == '__main__':
    if len(sys.argv) != 3:
        print_usage_and_exit()
    models = {}
    for g in gigasenses:
        models[g] = load_model("../../models/seeds." + g + ".h5")
    nouns = get_set_from_file(sys.argv[1])
    for n in nouns:
        occ[n] = 0
        num[n] = {}
        denum[n] = {}
        for g in gigasenses:
            num[n][g] = 0
            denum[n][g] = 0
    directory = "../contexts"
    if not os.path.exists(directory):
        os.makedirs(directory)
    mcf_filename = sys.argv[2]
    result_file = io.open(directory + "/" + os.path.basename(mcf_filename[:-4]), 'w', encoding='utf8')
    file = io.open(mcf_filename, "r", encoding='utf8')
    buffer = ""
    for line in file:
        tmp = line.split("\t")
        if len(tmp) == 7:
            buffer += line
            if tmp[6] == "1\n":
                update(buffer, nouns, result_file)
                buffer = ''
            if len(nouns) - len(nouns_finished) == 0:
                break

