#!/usr/bin/env python3

from keras.models import Sequential
from keras.layers import Dense, Activation, Dropout
import tensorflow as tf
import numpy as np
import sys
import argparse

parser = argparse.ArgumentParser(description="Perform WSD using supervised MLP")

parser.add_argument("--train", required=True,
                    help="The training set in TSV with nouns in 2nd column and supersense in 3rd column",
                    type=argparse.FileType(mode='r',encoding='utf-8'))
parser.add_argument("--dev", default=None,
                    help="The dev set in TSV with nouns in 2nd column and supersense in 3rd column",
                    type=argparse.FileType(mode='r',encoding='utf-8'))
parser.add_argument("--test", required=True,
                    help="The test set in TSV with nouns in 2nd column and supersense in 3rd column",
                    type=argparse.FileType(mode='r',encoding='utf-8'))

args = parser.parse_args()

########################################################################

def readData(fileName):
    try:
        fic = fileName
    except IOError:
        print("File", fileName, "does not exist")
        return None

    lx = []
    ly = []
    for ligne in fic:
        ligne = ligne.strip('\n\r')
        x = ligne.split()
        x.pop(0) #sentence id
        x.pop(0) # word
        supersense = int(x.pop(0))
      
        x_float = [float(elt) for elt in x]
        lx.append(x_float)

        y = [0] * 6
        y[supersense - 1] = 1
        ly.append(y)
    fic.close()

    x_formatted = np.array(lx, dtype="float")
    y_formatted = np.array(ly, dtype="int")
    print(len(x_formatted), "exemples lus")
    #print(x_train)
    return (x_formatted, y_formatted)

########################################################################

def argmax(l):
  i = 1
  max = l[0] 
  arg = 0
  while i < len(l) :
    if l[i] > max :
      max = l[i]
      arg = i
    i = i + 1
  return arg

########################################################################

def train(trainFilename, devFilename, testFilename):
    print("train =", trainFilename)
    (x_train, y_train) = readData(trainFilename)
    (x_dev, y_dev) = readData(devFilename)
    (x_test, y_test) = readData(testFilename)    

    for i in range(10):
        coucheCachee = Dense(units=2048, activation='relu', input_dim=12)
        coucheDecision = Dense(units=6, activation='softmax')
        model = Sequential()
        model.add(coucheCachee)
        model.add(Dropout(0.40))
        model.add(coucheDecision)
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        es = tf.keras.callbacks.EarlyStopping(
            monitor="val_accuracy",
            min_delta=0,
            patience=100,
            verbose=0,
            mode="auto",
            baseline=None,
            restore_best_weights=True,
        )

        model.fit(x_train, y_train, epochs=40, batch_size=100, validation_split = 0.1, verbose=0)
#        model.fit(x_train, y_train, epochs=40, batch_size=100, validation_split = 10, verbose=1)
#        model.fit(x_train, y_train, epochs=40, batch_size=100, validation_data=(x_dev, y_dev), verbose=1, callbacks=[es])

        score = model.evaluate(x_test, y_test)
        print(testFilename.name, i, score[1])
        score = model.evaluate(x_dev, y_dev)
        print(devFilename.name, i, score[1])
#        model.save("modele.bin")

#        l_pred = model.predict(x_test, batch_size=None, verbose=0, steps=None)
#        i = 0
#        ok = 0
#        while i < len(l_pred) :
#            print(l_pred[i])
#            predicted = argmax(l_pred[i])
#            print(predicted)
#            gold = argmax(y_test[i])
#            print('pred =', predicted, 'gold =', gold)
#            if(gold == predicted):
#                ok += 1
#            i = i + 1

#        print('accuracy = ', ok / i)

########################################################################

train(args.train, args.dev, args.test)
