#!/usr/bin/env python3

import sys
import argparse
import re
import pdb

parser = argparse.ArgumentParser(description="Perform WSD based on most frequent supersense")

parser.add_argument("--train", required=True,
                    help="The training set in TSV with nouns in 2nd column and supersense in 3rd column",
                    type=argparse.FileType(mode='r',encoding='utf-8'))
parser.add_argument("--dev", default=None,
                    help="The dev set in TSV with nouns in 2nd column and supersense in 3rd column",
                    type=argparse.FileType(mode='r',encoding='utf-8'))
parser.add_argument("--test", required=True,
                    help="The test set in TSV with nouns in 2nd column and supersense in 3rd column",
                    type=argparse.FileType(mode='r',encoding='utf-8'))
parser.add_argument("--quiet", default=False,
                    help="Does not output the predictions on stdout, only the accuracy on stderr",
                    action="store_true")


args = parser.parse_args()

lexicon = {}
supersenses = {}

with args.train as trainfile :
  for trainline in trainfile :
    fields = trainline.strip().split("\t")
    noun = fields[1]
    supersense = fields[2]
    supersensedict = lexicon.get(noun,{})
    supersensedict[supersense] = supersensedict.get(supersense, 0) + 1
    lexicon[noun] = supersensedict
    supersenses[supersense] = supersenses.get(supersense, 0) + 1  
    
defaultsupersense = max(supersenses, key=lambda key: supersenses[key])

print("Stats of the training corpus:", file=sys.stderr)
n = .01 * sum(supersenses.values())
for ss in sorted(supersenses.items()) :
  print("  * {} -> {} ({:.2f}%)".format(ss[0],ss[1],ss[1]/n), file=sys.stderr) 

supersenses = {}
accuracy = 0
totalnouns = 0
backedoff = 0
with args.test as testfile :
  for testline in testfile :
    totalnouns += 1
    fields = testline.strip().split("\t")
    noun = fields[1]
    refsupersense = fields[2]
    supersenses[refsupersense] = supersenses.get(refsupersense, 0) + 1  
    supersensedict = lexicon.get(noun,{})
    if supersensedict :
      predsupersense = max(supersensedict, key=lambda key: supersensedict[key])
    else :
      predsupersense = defaultsupersense
      backedoff += 1
    #if len(supersensedict) > 1 :
    #  pdb.set_trace()
    if predsupersense == refsupersense :
      accuracy += 1
    fields[2] = predsupersense
    if not args.quiet :
      print("\t".join(fields))
acc_templ = """Test contains {} nouns, of which:
  * {} are absent in train and backed off ({:.4f})
  * {} are correctly predicted ({:.4f})"""
print(acc_templ.format(totalnouns, backedoff, 1.0 * backedoff / totalnouns,
                       accuracy, 1.0 * accuracy / totalnouns), file=sys.stderr)
  
print("Stats of the test corpus:", file=sys.stderr)
n = .01 * sum(supersenses.values())
for ss in sorted(supersenses.items()) :
  print("  * {} -> {} ({:.2f}%)".format(ss[0],ss[1],ss[1]/n), file=sys.stderr) 


